import getpass
import platform
from selenium import webdriver
from selenium.webdriver.common.by import By


BROWSER_OPERATION_TIMEOUT = 60


WINDOWS_UA = {
    'user_agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36',
    'disable_ch': True,
    'platform': 'Windows',
    'mobile': False,
}

LINUX_UA = {
    'user_agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36',
    'disable_ch': True,
    'platform': 'Linux',
    'mobile': False,
}

MACOS_UA = {
    'user_agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 13_3_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36',
    'disable_ch': True,
    'platform': 'macOS',
    'mobile': False,
}


def start_chrome(profile_dir=None, socks5_proxy=None, proxy=None, size=(1366, 768), position=(0, 0), user_agent=None, mobile_emulation=None, timeout=None):
    if timeout is None:
        webdriver.remote.remote_connection.RemoteConnection.set_timeout(BROWSER_OPERATION_TIMEOUT)
    else:
        webdriver.remote.remote_connection.RemoteConnection.set_timeout(timeout)

    opt = webdriver.ChromeOptions()
    opt.add_experimental_option('excludeSwitches', [
        'enable-automation',
        'disable-background-networking',
        'disable-default-apps',
        'disable-hang-monitor',
        'enable-blink-features',
        'no-service-autorun',
        'test-type',
        'use-mock-keychain',
    ])
    # Disable remember password feature
    opt.add_experimental_option('prefs', {
        'credentials_enable_service': False,
        'profile.password_manager_enabled': False,
    })
    # Disable browser notification
    opt.add_argument('--disable-notifications')
    # Disable default browser check and confirm
    opt.add_argument('--no-default-browser-check')
    # Disable auto update check
    opt.add_argument("--simulate-outdated-no-au='Tue, 31 Dec 2099 23:59:59 GMT'")
    # If in linux and user is root just add --no-sandbox
    if platform.system() == 'Linux' and getpass.getuser() == 'root':
        opt.add_argument('--no-sandbox')

    if mobile_emulation is not None:
        user_agent = None
        opt.add_experimental_option("mobileEmulation", mobile_emulation)

    # user_agent is string, set user agent as chrome parameter and disable SEC-CH-UA
    if user_agent is not None:
        if isinstance(user_agent, str):
            opt.add_argument('--user-agent=%s' % user_agent)
            opt.add_argument('--disable-features=UserAgentClientHint')
        elif isinstance(user_agent, dict):
            opt.add_argument('--user-agent=%s' % user_agent['user_agent'])
            if user_agent.get('disable_ch', False):
                opt.add_argument('--disable-features=UserAgentClientHint')
    if position is not None:
        opt.add_argument('--window-position=%s,%s' % (position[0], position[1]))
    if size is not None:
        opt.add_argument('--window-size=%s,%s' % (size[0], size[1]))
    if profile_dir is not None:
        opt.add_argument('--user-data-dir=%s' % profile_dir)
    if socks5_proxy is not None:
        opt.add_argument('--proxy-server=socks5://%s' % socks5_proxy)
    if proxy is not None:
        opt.add_argument('--proxy-server=%s' % proxy)
    c = webdriver.Chrome(options=opt)
    # user_agent is dict, set user agent and SEC-CH-UA via CDP request
    if user_agent is not None and isinstance(user_agent, dict):
        if not user_agent.get('disable_ch', False):
            c.execute_cdp_cmd("Emulation.setUserAgentOverride", {
                'userAgent': user_agent['user_agent'],
                'platform': user_agent.get('platform', 'Windows'),
                'userAgentMetadata': {
                    'platform': user_agent.get('platform', 'Windows'),
                    'platformVersion': user_agent.get('version', ''),
                    'architecture': user_agent.get('architecture', ''),
                    'model': user_agent.get('model', ''),
                    'mobile': user_agent.get('mobile', False),
                },
            })
    return c


def get(b, url):
    b.get(url)


def __check(src, dst, mode):
    print(src, dst, mode)
    if mode == 'contains':
        return dst in src
    elif mode == 'equals':
        return dst == src
    return False


# Switch to browser tab
# mode: contains, equals
def switch_to_tab(driver, idx=None, name='', url='', mode='contains'):
    hdls = driver.window_handles
    try:
        current_hdl = driver.current_window_handle
    except Exception:
        current_hdl = hdls[0]

    if idx is not None:
        if idx >= len(hdls):
            return
        driver.switch_to.window(hdls[idx])

    found = False
    if name != '':
        for hdl in hdls:
            driver.switch_to.window(hdl)
            if __check(driver.title, name, mode):
                found = True
                break

        if not found:
            driver.switch_to.window(current_hdl)
        return

    if url != '':
        for hdl in hdls:
            driver.switch_to.window(hdl)
            if __check(driver.current_url, url, mode):
                found = True
                break

        if not found:
            driver.switch_to.window(current_hdl)
        return


def scroll_page(driver, steps):
    driver.execute_script("window.scrollTo(0, window.scrollY + %d)" % int(steps))


def create_new_tab(driver, url=''):
    driver.switch_to.new_window('tab')
    if url != '':
        driver.get(url)


def list_tabs(driver):
    ret = []
    data = driver.execute_cdp_cmd('Target.getTargets', {})
    targets = data.get('targetInfos', [])
    for handle in driver.window_handles:
        for tgt in targets:
            if tgt.get('targetId', '') == handle:
                ret.append(tgt)
    return ret


def snap_page(driver, fname=None):
    if fname is None:
        return driver.get_screenshot_as_png()
    return driver.get_screenshot_as_file(fname)


def get_parent(elem):
    return elem.find_element(By.XPATH, 'parent::*')


def __transform_by(kwargs):
    if len(kwargs) != 1:
        raise Exception('Invalid query argument, require only one query')

    by, query = list(kwargs.items())[0]
    if by.lower() == 'id':
        by = By.ID
    elif by.lower() == 'tag':
        by = By.TAG_NAME
    elif by.lower() == 'style':
        by = By.CLASS_NAME
    elif by.lower() == 'css':
        by = By.CSS_SELECTOR
    elif by.lower() == 'name':
        by = By.NAME
    elif by.lower() == 'xpath':
        by = By.XPATH
    else:
        raise Exception('Invalid query type: %s' % by)

    return by, query


def find_element(driver, **kwargs):
    by, query = __transform_by(kwargs)
    elems = driver.find_elements(by, query)
    if len(elems) == 0:
        return None
    return elems[0]


def find_elements(driver, **kwargs):
    by, query = __transform_by(kwargs)
    return driver.find_elements(by, query)



import time
import platform


def wait(sec=1):
    time.sleep(sec)


def wait_input(prompt='Input: '):
    return input(prompt)


def os_name():
    return platform.system()


def is_windows():
    return os_name() == 'Windows'


def is_linux():
    return os_name() == 'Linux'


def is_macos():
    return os_name() == 'Darwin'



import time
import platform
import pyautogui
import pyperclip


# message box
def alert(text, title='Alert'):
    pyautogui.alert(text=text, title=title, button='OK')


def confirm(text, title='Confirm'):
    ret = pyautogui.confirm(text=text, title=title, buttons=['OK', 'Cancel'])
    return ret == 'OK'


def prompt(text, title='Please Input', default=''):
    ret = pyautogui.prompt(text, title, default)
    return ret


def password(text, title='Please Input Password', default=''):
    ret = pyautogui.password(text, title, default)
    return ret


# clipboard and keyboard control
def clip(text):
    pyperclip.copy(text)


def snap(fname=None):
    if fname is None:
        return pyautogui.screenshot()
    return pyautogui.screenshot(fname)


def write(text, interval=0.4):
    pyautogui.write(text, interval=interval)


def hotkey(*args):
    pyautogui.hotkey(*args)
    time.sleep(0.4)


def paste(text=None):
    if text is not None:
        pyperclip.copy(text)
    if platform.system() == 'Darwin':
        hotkey('command', 'v')
    else:
        hotkey('ctrl', 'v')


# Mouse control
def move(x, y):
    pyautogui.moveTo(x, y, duration=0.2)


def click(x=None, y=None, clicks=1):
    if x is not None and y is not None:
        move(x, y)
    pyautogui.click(clicks=clicks)


def dclick(x=None, y=None):
    click(x, y, 2)


def rclick(x=None, y=None, clicks=1):
    if x is not None and y is not None:
        move(x, y)
    pyautogui.click(button='right', clicks=clicks)


def mdown(x=None, y=None):
    if x is not None and y is not None:
        move(x, y)
    pyautogui.mouseDown()


def mup(x=None, y=None):
    if x is not None and y is not None:
        move(x, y)
    pyautogui.mouseUp()


def scroll(steps):
    pyautogui.scroll(steps)


def click_and_paste(x, y, text):
    clip(text)
    move(x, y)
    click()
    paste()


def click_and_type(x, y, text, interval=0.2):
    move(x, y)
    click()
    write(text, interval)



import cv2
import pyautogui
import numpy as np
from .gui import move, click, paste
from .utils import wait


SCREEN_HRATIO = 1
SCREEN_WRATIO = 1
THRESHOLD = 0.8
ELEMENT_IMAGE_RATIO = (1, 1)


def init_img(element_ratio=None):
    global SCREEN_HRATIO
    global SCREEN_WRATIO
    global ELEMENT_IMAGE_RATIO
    img = pyautogui.screenshot()
    w, h = img.size
    ss = pyautogui.size()
    SCREEN_HRATIO = h / ss.height
    SCREEN_WRATIO = w / ss.width
    if element_ratio is not None:
        ELEMENT_IMAGE_RATIO = element_ratio
    print('Screen Ratio: %s, %s' % (SCREEN_WRATIO, SCREEN_HRATIO))
    print('Elememt Image Ratio: %s, %s' % ELEMENT_IMAGE_RATIO)


# Scale image to new ratio
def scale_image(img, ratio):
    xr, yr = ratio[0], ratio[1]
    if xr == 1 and yr == 1:
        return img

    shape = img.shape
    if len(shape) < 2:
        return img

    h, w = shape[0], shape[1]
    nw = int(w * xr)
    nh = int(h * yr)
    op = cv2.INTER_LINEAR
    if nw < xr and nh < yr:
        # Shrink better
        op = cv2.INTER_AREA
    return cv2.resize(img, (nw, nh), interpolation=op)


def __global_scale_image(img):
    return scale_image(img, ELEMENT_IMAGE_RATIO)


def find_image_element(element, pmode='center', debug=False, threshold=None):
    elem_img = __global_scale_image(cv2.imread(element))
    h, w, _ = elem_img.shape
    pil_screen_img = pyautogui.screenshot()
    screen_img = cv2.cvtColor(np.array(pil_screen_img), cv2.COLOR_RGB2BGR)

    src_img_g = cv2.cvtColor(screen_img, cv2.COLOR_BGR2GRAY)
    elem_img_g = cv2.cvtColor(elem_img, cv2.COLOR_BGR2GRAY)

    result = cv2.matchTemplate(src_img_g, elem_img_g, cv2.TM_CCOEFF_NORMED)
    (min_val, max_val, min_loc, max_loc) = cv2.minMaxLoc(result)
    x, y = max_loc[0], max_loc[1]

    if debug:
        print('DEBUG: x = %s, y = %s, max_val = %s' % (x, y, max_val))
        # Draw element rectangle
        cv2.rectangle(screen_img, (x, y), (x + w, y + h), (0, 0, 255), 2)
        cv2.imwrite('debug.png', screen_img)

    if threshold is None:
        threshold = THRESHOLD

    if max_val <= threshold:
        return False, 0, 0

    if pmode == 'center':
        x = x / SCREEN_WRATIO + w / 2 / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO
    elif pmode == 'topleft':
        x = x / SCREEN_WRATIO
        y = y / SCREEN_HRATIO
    elif pmode == 'topright':
        x = x / SCREEN_WRATIO + w / SCREEN_WRATIO
        y = y / SCREEN_HRATIO
    elif pmode == 'centerleft':
        x = x / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO
    elif pmode == 'centerright':
        x = x / SCREEN_WRATIO + w / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO

    return True, int(x), int(y)


def find_image_element2(location, element, pmode='center', debug=False, threshold=None):
    location_img = __global_scale_image(cv2.imread(location))
    lh, lw, _ = location_img.shape
    pil_screen_img = pyautogui.screenshot()
    screen_img = cv2.cvtColor(np.array(pil_screen_img), cv2.COLOR_RGB2BGR)

    src_img_g = cv2.cvtColor(screen_img, cv2.COLOR_BGR2GRAY)
    loc_img_g = cv2.cvtColor(location_img, cv2.COLOR_BGR2GRAY)
    result = cv2.matchTemplate(src_img_g, loc_img_g, cv2.TM_CCOEFF_NORMED)
    (min_val, max_val, min_loc, max_loc) = cv2.minMaxLoc(result)

    if debug:
        print('DEBUG: x = %s, y = %s, max_val = %s' % (max_loc[0], max_loc[1], max_val))

    if threshold is None:
        threshold = THRESHOLD

    if max_val <= threshold:
        return False, 0, 0

    loc_x, loc_y = max_loc[0], max_loc[1]
    elem_img = __global_scale_image(cv2.imread(element))
    h, w, _ = elem_img.shape
    elem_img_g = cv2.cvtColor(elem_img, cv2.COLOR_BGR2GRAY)
    result = cv2.matchTemplate(loc_img_g, elem_img_g, cv2.TM_CCOEFF_NORMED)
    (min_val, max_val, min_loc, max_loc) = cv2.minMaxLoc(result)

    x, y = max_loc[0], max_loc[1]
    x += loc_x
    y += loc_y
    if debug:
        print('DEBUG: x = %s, y = %s, max_val = %s' % (x, y, max_val))
        # Draw location rectangle
        cv2.rectangle(screen_img, (loc_x, loc_y), (loc_x + lw, loc_y + lh), (0, 0, 255), 2)
        # Draw element rectangle
        cv2.rectangle(screen_img, (x, y), (x + w, y + h), (0, 0, 255), 2)
        cv2.imwrite('debug.png', screen_img)

    if max_val <= threshold:
        return False, 0, 0

    if pmode == 'center':
        x = x / SCREEN_WRATIO + w / 2 / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO
    elif pmode == 'topleft':
        x = x / SCREEN_WRATIO
        y = y / SCREEN_HRATIO
    elif pmode == 'topright':
        x = x / SCREEN_WRATIO + w / SCREEN_WRATIO
        y = y / SCREEN_HRATIO
    elif pmode == 'centerleft':
        x = x / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO
    elif pmode == 'centerright':
        x = x / SCREEN_WRATIO + w / SCREEN_WRATIO
        y = y / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO

    return True, int(x), int(y)


def click_image(elem, debug=False, threshold=None):
    found, x, y = find_image_element(elem, debug=debug, threshold=threshold)
    if not found:
        return

    move(x, y)
    click()


def click_image2(location, elem, debug=False, threshold=None):
    found, x, y = find_image_element2(location, elem, debug=debug, threshold=threshold)
    if not found:
        return
    move(x, y)
    click()


def exists_image(elem, debug=False, threshold=None):
    found, _, _ = find_image_element(elem, debug=debug, threshold=threshold)
    return found


def exists_image2(location, elem, debug=False, threshold=None):
    found, _, _ = find_image_element2(location, elem, debug=debug, threshold=threshold)
    return found


def wait_untile_exists(elem, next_op=None, max_wait=20, duration=1, debug=False, threshold=None):
    found = False
    for i in range(max_wait):
        found, x, y = find_image_element(elem, debug=debug, threshold=threshold)
        if found:
            break
        wait(duration)

    if found:
        if next_op == 'click':
            move(x, y)
            click()
        elif next_op == 'hover':
            move(x, y)
        elif next_op == 'click_paste':
            move(x, y)
            click()
            paste()
        return x, y
    return None, None


def wait_untile_exists2(location, elem, next_op=None, max_wait=20, duration=1, debug=False, threshold=None):
    found = False
    for i in range(max_wait):
        found, x, y = find_image_element2(location, elem, debug=debug, threshold=threshold)
        if found:
            break
        wait(duration)

    if found:
        if next_op == 'click':
            move(x, y)
            click()
        elif next_op == 'hover':
            move(x, y)
        elif next_op == 'click_paste':
            move(x, y)
            click()
            paste()
        return x, y
    return None, None



from .gui import alert, confirm, prompt, clip, snap, write, hotkey, paste, move, click, dclick, rclick, mdown, mup, scroll, click_and_paste, click_and_type
from .images import init_img, find_image_element, find_image_element2, click_image, click_image2, exists_image, exists_image2, wait_untile_exists, wait_untile_exists2
from .utils import wait, wait_input, os_name, is_windows, is_linux, is_macos
from .chrome import start_chrome, switch_to_tab, create_new_tab, list_tabs, scroll_page, snap_page, WINDOWS_UA, LINUX_UA, MACOS_UA, find_element, find_elements, get_parent
from .ocr import init_ocr, find_ocr_element


# Initialize
def init(screen_ratio=(1, 1), enable_ocr=False, ocr_langs=['en'], enable_gpu=False):
    init_img(screen_ratio)
    if enable_ocr:
        init_ocr(ocr_langs, enable_gpu)


# Browser
def chrome(profile_dir=None, socks5_proxy=None, proxy=None, size=(1266, 800), position=(0, 0), user_agent=None, mobile_emulation=None, timeout=None):
    return start_chrome(profile_dir, socks5_proxy, proxy, size, position, user_agent, mobile_emulation, timeout)


def switch_tab(driver, idx=None, name='', url='', mode='contains'):
    switch_to_tab(driver, idx, name, url, mode)


def new_tab(driver, url=''):
    create_new_tab(driver, url)


def tabs(driver):
    return list_tabs(driver)


def get_new_tabs(driver, origin_tabs):
    new_tabs = list_tabs(driver)
    if len(new_tabs) <= len(origin_tabs):
        return []

    origin_ids = [item.get('targetId', '') for item in origin_tabs]

    ret = []
    for tab in new_tabs:
        if tab['targetId'] not in origin_ids:
            ret.append(tab)
    return ret


def shell(local=None, banner='PyAutoBot Shell'):
    import code
    import importlib
    from selenium.webdriver.common.by import By
    # Setup rpa module
    spec = importlib.machinery.ModuleSpec('bot', None)
    mbot = importlib.util.module_from_spec(spec)
    for fn in __all__:
        setattr(mbot, fn, globals()[fn])

    # Default local should have rpa and By module
    local_obj = {'bot': mbot, 'By': By}
    if isinstance(local, dict):
        local_obj.update(local)
    code.interact(banner=banner, local=local_obj)


__all__ = [
    # Initialize functions
    'init', 'init_img', 'init_ocr',
    # OS Related functions
    'os_name', 'is_windows', 'is_linux', 'is_macos',
    # Wait and sleep functions
    'wait', 'wait_input', 'shell',
    # Prompt and dialog functions
    'alert', 'confirm', 'prompt',
    # Copy/paste and keyboard input
    'clip', 'write', 'hotkey', 'paste', 'click_and_paste', 'click_and_type',
    # Mouse control functions
    'move', 'click', 'dclick', 'rclick', 'mdown', 'mup', 'scroll',
    # Image find functions
    'find_image_element', 'find_image_element2',
    'click_image', 'click_image2',
    'exists_image', 'exists_image2',
    'wait_untile_exists', 'wait_untile_exists2',
    # Chrome related functions
    'chrome', 'switch_tab', 'new_tab', 'tabs', 'get_new_tabs', 'scroll_page',
    'WINDOWS_UA', 'LINUX_UA', 'MACOS_UA',
    'find_element', 'find_elements', 'get_parent',
    # Screen shot or browser screen shot
    'snap', 'snap_page',
    # OCR functions
    'find_ocr_element',
]



import cv2
import pyautogui
import numpy as np


EASYOCR_READER = None


def init_ocr(langs=['en'], enable_gpu=False):
    import easyocr
    global EASYOCR_READER
    EASYOCR_READER = easyocr.Reader(langs, gpu=enable_gpu)


def compare_text(text, rtext):
    if text in rtext:
        return True
    elif text.replace(' ', '') in rtext.replace(' ', ''):
        return True
    return False


def find_ocr_element(text, pmode='center', rect=None, debug=True, threshold=None):
    from .images import THRESHOLD, SCREEN_HRATIO, SCREEN_WRATIO
    print(SCREEN_HRATIO, SCREEN_WRATIO)
    if EASYOCR_READER is None:
        print('Alert: OCR system not initialized')
        return False, 0, 0

    if threshold is None:
        threshold = THRESHOLD

    pil_screen_img = pyautogui.screenshot()
    screen_np = np.array(pil_screen_img)
    screen_img = cv2.cvtColor(screen_np, cv2.COLOR_RGB2BGR)
    offset_x, offset_y = 0, 0
    if rect is not None and len(rect) == 4:
        x, y, w, h = rect
        offset_x = x
        offset_y = y
        cropped = screen_np[y:y + h, x:x + w]
        result = EASYOCR_READER.readtext(cropped)
    else:
        result = EASYOCR_READER.readtext(screen_np)
    founds = []
    confi_max = 0
    found_max = None
    for line in result:
        box, rtext, confi = line[0], line[1], line[2]
        if len(box) < 4 or confi < threshold:
            continue

        if compare_text(text, rtext):
            x, y = box[0][0], box[0][1]
            x2, y2 = box[-2][0], box[-2][1]
            w = int(x2 - x)
            h = int(y2 - y)
            item = [(int(x), int(y), w, h), rtext, confi]
            founds.append(item)
            if confi > confi_max:
                confi_max = confi
                found_max = item

    if debug:
        for f in founds:
            x, y, w, h = f[0]
            print('DEBUG: x = %s, y = %s, confidence = %s, text = %s' % (x, y, f[2], f[1]))
            cv2.rectangle(screen_img, (x + offset_x, y + offset_y), (x + w + offset_x, y + h + offset_y), (0, 0, 255), 2)
            cv2.imwrite('debug.png', screen_img)

    if found_max is None:
        return False, 0, 0

    x, y, w, h = found_max[0]
    if pmode == 'center':
        x = (x + offset_x) / SCREEN_WRATIO + w / 2 / SCREEN_WRATIO
        y = (y + offset_y) / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO
    elif pmode == 'topleft':
        x = (x + offset_x) / SCREEN_WRATIO
        y = (y + offset_y) / SCREEN_HRATIO
    elif pmode == 'centerleft':
        x = (x + offset_x) / SCREEN_WRATIO
        y = (y + offset_y) / SCREEN_HRATIO + h / 2 / SCREEN_HRATIO

    return True, int(x), int(y)



from setuptools import setup


dependence = [
    'opencv-python',
    'pyautogui',
    'pyperclip',
    'selenium',
    'easyocr',
    'sympy',
    'lazy_loader',
]

desc = 'pyautobot is a Python library provide functions that do robotic process automation jobs.'

long_desc = '''pyautobot is a Python library provide functions that do robotic process automation jobs.

It combines below librarys to archive the target:

* PyAutoGUI and pyperclip: control clipboard, mouse and keyboard operations.
* Selenium: control Chrome browser.
* OpenCV-Python: provide graphic based operation.
* EasyOCR: provide OCR functions.
'''

version = '1.0.3'

setup(
    name='pyautobot',
    version=version,
    url='https://github.com/blacktear23/pyautobot',
    author='blacktear23',
    author_email='blacktear23@gmail.com',
    description=desc,
    long_description=long_desc,
    long_description_content_type="text/markdown",
    license='BSD',
    packages=['pyautobot'],
    scripts=['bin/pyautobot'],
    test_suite='tests',
    install_requires=dependence,
    keywords="automation rpa",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)
